//-----------------------------------------------------------------------------
// Температурный сервер на базе МК ATMega328P/Arduino Pro Mini etc и датчиков
// температуры DS18B20
//-----------------------------------------------------------------------------
// Файл main.c
// OldBean, 16.06.2017
//-----------------------------------------------------------------------------
// Подробнее про LITE и no LITE (с индикатором и кнопкой) см. соотв. топик
//-----------------------------------------------------------------------------
#include <avr/io.h>
#include <compat/twi.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <util/atomic.h>
//-----------------------------------------------------------------------------
//#define LITE // Если вариант LITE - раскомментируйте эту строку
#define ADDR 0x07 	// Адрес сервера на шине I2C
//-----------------------------------------------------------------------------
#define DEBUG
#ifdef DEBUG
uint8_t ledMask = 0b00001000; // Светодиод для отладки висит на пине PC3
#endif
//-----------------------------------------------------------------------------
// Прототипы функций
//-----------------------------------------------------------------------------
uint32_t millis(void);          // Аналог функции Arduino
uint8_t w1_reset();             // Сброс всех шин 1-Wire, обнаружение датчиков
void w1_write_byte(uint8_t b);  // Отсылаем один и тот же байт на все шины
void w1_read_bytes(uint8_t *b); // Чтение байтов со всех шин 1-Wire
#ifndef LITE // Дополнительный функционал для работы с индикатором и кнопкой
void pinMode(uint8_t dig, uint8_t dir);       // Упрощенный аналог ф-и Arduino
void digitalWrite(uint8_t dig, uint8_t val);  // Упрощенный аналог ф-и Arduino
void ShowSegments();            // Подсвечивание очередной группы сегментов
uint8_t GetKey();               // Возвращает состояние клавиатуры
void UpdateDigits(int val);     // Подготовка данных (цифр) для визуализации
void ShowSegments();            // Подсвечивание очередной группы сегментов
#endif
//-----------------------------------------------------------------------------
// Тактирование процессов
//-----------------------------------------------------------------------------
volatile uint32_t msc32 = 0; // Счетчик миллисекунд
//------------------------------------------------------------------------------
uint32_t millis(void) { // Аналог функции Arduino
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {return msc32;}
}
//------------------------------------------------------------------------------
ISR(TIMER2_COMPA_vect) { // Обработчик прерывания таймера TIMER2 (раз в 1 мс)
  msc32++; // Счетчик мс
#ifndef LITE  // Низкоприоритетный фрагмент для работы с индикатором
  sei();
  if(msc32 & 0x00000001) // Каждые 2 мс - посветка следующей группы сегментов
    ShowSegments();
#endif
}
//-----------------------------------------------------------------------------
// Констанды и переменные для работы с датчиками DS18B20
//-----------------------------------------------------------------------------
#ifndef LITE  // Расширенный вариант (с индикатором и кнопкой)
uint8_t mask1 = 0b00001111; // Доступны только 4 младших бита порта D
uint8_t mask0 = 0b11110000; // mask0 = ~mask1
#else         // Вариант LITE
uint8_t mask1 = 0b11111111; // Все 8 пинов доступны для шин 1-Wire
uint8_t mask0 = 0b00000000; // mask0 = ~mask1
#endif
#define W1_READY      0     // Температурные данные готовы
#define W1_CONVERSION 1     // Датчики в процессе преобразования
static uint8_t w1Status = W1_READY;
static uint8_t devs;        // Маска наличия (1) датчиков

#ifndef LITE  // Расширенный вариант (с индикатором и кнопкой)
uint8_t curTS = 0;  // Номер текущего датчика. Его температура выводится на 
                    // индикатор. Совпадает с номером пина порта D к которому
                    // подключена соответствующая шина 1-Wire
#endif

// Буфер со значениями температур. Каждая пара содержит код температуры
// (старший и младщий байт). Значение температуры в °C получается умножением
// этого кода на величину кванта 0.0625.
uint8_t buf[16] = {   1,  82,   2, 165, 255, 255,   4, 226, // Числа - для отладки
                      0,   0,   0,   0,   0,   0,   0,   0};
uint8_t bufL = 16;
uint8_t index = 0;
//-----------------------------------------------------------------------------
// Реализация функций работы с шинами 1-Wire
//-----------------------------------------------------------------------------
uint8_t w1_reset() { // Сброс всех шин 1-Wire, обнаружение датчиков
//  cli(); // Запретим прерывания
  DDRD |= mask1; // Пины шин - на вывод
  PORTD &= mask0; // Установим низкий уровень на всех шинах 1-Wire
  _delay_us(480);
  DDRD &= mask0; // Пины шин - на ввод
  _delay_us(60);
  uint8_t devs = ~(PIND | mask0); // Там, где есть датчик - будет 1
  _delay_us(420); //ждем оставшееся время до 480мкс
  return devs;
//  sei(); // Разрешаем прерывания
}
//-----------------------------------------------------------------------------
void w1_write_byte(uint8_t b) { // Отсылаем один и тот же байт на все шины
  uint8_t mask = 0b00000001;
  for(uint8_t bit = 0; bit < 8; bit++) {
    cli(); // Запретим прерывания
    DDRD |= mask1; // Пины шин - на вывод
    PORTD &= mask0; // 0 на всех шинах 1-Wire
    _delay_us(1);
    if(b & mask) // Шлем единичку
      DDRD &= mask0; // "Отпустим" шины - переводем порт в режим чтения
    _delay_us(59);
    DDRD &= mask0;
    _delay_us(5);
    mask <<= 1;          
    sei(); // Разрешаем прерывания
  }
}
//-----------------------------------------------------------------------------
void w1_read_bytes(uint8_t *b) { // Чтение байтов со всех шин 1-Wire
  uint8_t wb[8]; // Вспомогательный массив, который заносятся читаемые биты.
                 // Каждый байт wb содержит соответствующие биты всех датчиков.
                 // Например, wb[3] содержит 3-и биты всех датчиков.
  cli(); // Запретим прерывания
  for(uint8_t bit = 0; bit < 8; bit++) {
    DDRD |= mask1; // Пины шин - на вывод
    PORTD &= mask0; // 0 на всех шинах 1-Wire
    _delay_us(1);
    DDRD &= mask0; // Пины шин - на ввод
    _delay_us(14);
    wb[bit] = PIND & mask1; // Читаем все биты с номером bit
    _delay_us(45);
  }
  _delay_us(5);
  // Преобразуем полученную информацию (массив wb) к нормальной форме, чтобы
  // каждый байт содержал все биты, полученные от одного датчика. По сути, 
  // нам нужно просто транспонировать матрицу 8x8, представленную массивом wb.
  uint8_t bmask = 0b00000001;
  for(uint8_t bit = 0; bit < 8; bit++) {
    uint8_t nmask = 0b00000001;
    for(uint8_t sn = 0; sn < 8; sn++) {
      if(devs & nmask) { // Датчик на этой шине есть 
        if(wb[bit] & nmask) b[sn] |= bmask; // Установим бит
        else b[sn] &= ~bmask; // Сбросим бит
      } else b[sn] |= bmask; // Если на шине нет датчика - будут все 1
      nmask <<= 1;
    }
    bmask <<= 1;
  }
  sei(); // Разрешаем прерывания
}

uint8_t gFlag = 0; // Флаг готовности температурных данных
//-----------------------------------------------------------------------------
ISR(TWI_vect) { // Обработчик прерываний TWI

#ifdef DEBUG
if(PORTC & ledMask) PORTC &= ~ledMask;
else PORTC |= ledMask;
#endif

  unsigned char status = TWSR & 0b11111000; // Статус - в старших 5 битах
//  cli(); // Запрещаем прерывания
  switch(status) { // Анализируем статус TWI
// В данном релизе ничего не читаем...
//    case TW_SR_SLA_ACK: // 0x60
//    case TW_SR_DATA_ACK:    // 0x80
//    case TW_SR_STOP: // 0xA0
//      break;
    case TW_ST_SLA_ACK: // 0xA8
      index = 0;
      TWDR = buf[index];
      break;
    case TW_ST_DATA_ACK:    // 0xB8
      if(++index < bufL) {
        TWDR = buf[index];
        if(index == bufL - 1) // Последний байт
          TWCR &= ~(1<<TWEA); // Не работает! Разобраться...
      } else TWDR = 255;
      break;
    case TW_ST_DATA_NACK:   // 0xC0: данные отправлены, принят NACK
    case TW_ST_LAST_DATA:   // 0xC8: последний байт отправлен, ACK принят
      TWCR |= (1<<TWEA);
      break;
    default:
      break;
  }
  TWCR |= (1<<TWINT); // Очищаем флаг TWINT
//  sei(); // Разрешаем прерывания
}
//-----------------------------------------------------------------------------
#ifndef LITE  // Расширенный вариант (с индикатором и кнопкой)
//#define INDICATOR_TEST  // Тестирование индикатора (для отладки)
//#define BUTTON_TEST     // Тестирование кнопки (для отладки)
#endif

#ifndef LITE  // Расширенный вариант (с индикатором и кнопкой)
//-----------------------------------------------------------------------------
// Константы и переменные для работы с индикатором
//-----------------------------------------------------------------------------
#define COMMON_CATHODE  // Если индикатор с общим анодом - необходимо
                        // закомментировать эту строку
#define DIGS 4          // Количество разрядов в индикаторе
#define DP_POS 1        // Положение десятичной точки. Точка не будет
                        // показываться если DP_POS >= DIGS
#ifdef COMMON_CATHODE 	// Для индикаторов с общим катодом
#define L_HIGH 1
#define L_LOW  0
#else                 	// Для индикаторов с общим анодом
#define L_HIGH 0
#define L_LOW  1
#endif
// Пины, к которым подключены сегменты индикатора (нумерация пинов
// совпадает с нумерацией пинов Arduino Pro Mini)
static uint8_t segPins[8] = {11, 15, 6, 8, 9, 12, 5, 7};
// Пины, к которым подключены общие катоды/аноды
static uint8_t digPins[4] = {10, 13, 14, 4};
static uint8_t digit[20] = { // Битовые маски цифр и некоторых знаков
							0b00111111, 0b00000110, 0b01011011, 0b01001111, 0b01100110,
							0b01101101, 0b01111101, 0b00000111, 0b01111111, 0b01101111,
							0b00000000, // Пробел
							0b00001000, 0b01000000, // Минус 
							0b00000001};
static int space = 10, minus = 12; // Вн. коды пары символов
static int seg = 0; // Номер текущего сегмента индикатора
static int   dpPos = DIGS + 1; // Текущее положение десятичной точки
static int maxval; // Максимальное число, отбражаемое на индикаторе + 1
static int d[DIGS]; // Цифры, показываемые в разрядах индикатора
static uint8_t mask = 0x01; // Маска для анализа выводимых цифр

static int val = 0; // Просто рабочая переменная, отображаемая на индикаторе

#define SHOW_TEMP   0 // Показ температуры текущего датчика
#define SHOW_TS_NUM 1 // Показ номера текущего датчика
static uint8_t dMode = SHOW_TEMP; // Текущий режим индикации
//-----------------------------------------------------------------------------
// Константы и переменные для работы с кнопкой
//-----------------------------------------------------------------------------
static uint8_t b_pin = 2; // Номер пина порта C, к которому подключена кнопка
static uint8_t bpm;       // Маска для этого пина (инициализируется в main :)

#define NO_ACT			    0           // С кнопкой ничего нового не происходит
#define PLUS_CLICK		  0b00000001  // Короткое нажатие кнопки
#define PLUS_PRESS		  0b00000100  // Удержание кнопки
#define PLUS_PRESS_REL	0b00010000	// Отпускание кнопки после удержания

static uint8_t bos = NO_ACT;  // Предыдущее состояние кнопок
static uint32_t tpr; 	        // Относительный момент нажатия кнопки в мс
static int dtBounce = 30;     // Задержка (мс) для устранения дребезга
static int dtPress = 400;     // Задержка, после которой считается, что кнопка
                              // удерживается в нажатом состоянии
//-----------------------------------------------------------------------------
#define INPUT  0              // Константы для совместимости с Arduino
#define OUTPUT 1
//-----------------------------------------------------------------------------
// Реализация функций -  аналогов ф-й Arduino
//-----------------------------------------------------------------------------
void pinMode(uint8_t dig, uint8_t dir) { // Упрощенный аналог ф-и Arduino
  volatile uint8_t *ddrReg; // Указатель на регистр порта
  uint8_t pin;
  if(dig <= 7) {ddrReg = &DDRD; pin = dig;}
  else if(dig <= 13) {ddrReg = &DDRB; pin = dig - 8;}
       else {ddrReg = &DDRC; pin = dig - 14;}
  uint8_t mask = 0x01 << pin;
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    if(dir) (*ddrReg) |= mask;  // Пин - на вывод
    else (*ddrReg) &= ~mask;    // Пин - на ввод
  }
}
//-----------------------------------------------------------------------------
void digitalWrite(uint8_t dig, uint8_t val) { // Упрощенный аналог ф-и Arduino
  volatile uint8_t *portReg; // Указатель на регистр порта
  uint8_t pin;
  if(dig <= 7) {portReg = &PORTD; pin = dig;}
  else if(dig <= 13) {portReg = &PORTB; pin = dig - 8;}
       else {portReg = &PORTC; pin = dig - 14;}
  uint8_t mask = 0x01 << pin;
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    if(val) (*portReg) |= mask; // Выставляем бит
    else (*portReg) &= ~mask; // Сбрасываем бит
  }
}
//-----------------------------------------------------------------------------
// Реализация функций индикатора и кнопки
//-----------------------------------------------------------------------------
void UpdateDigits(int val) { // Подготовка данных (цифр) для визуализации
  if(val >= maxval) {
    for(int i = 0; i < DIGS; i++) d[i] = minus; 
    return;
  }
  int wv = val, dv = maxval/10;
  for(int i = 0; i < DIGS; i++) {
    d[i] = wv/dv; wv -= d[i]*dv; dv /= 10;
  }
  int lim = (dpPos < DIGS)? dpPos : DIGS - 1;
  for(int i = 0; i < lim; i++) { // Убираем ведущие нули
    if(d[i] == 0) d[i] = space;
    else break;
  }
}
//-----------------------------------------------------------------------------
void ShowSegments() { // Подсвечивание очередной группы сегментов
  digitalWrite(segPins[seg], L_LOW);  // "Роняем" потенциал анода
                                      // "отработавшего" сегмента
  if(++seg < 8) mask <<= 1; // Переходим к следующему сегменту
  else { // Дошли до последнего сегмента
    seg = 0; mask = 0x01;
  }
  digitalWrite(segPins[seg], L_HIGH); // "Подбрасываем" потенциал анода нового
                                      // текущего сегмента
  for(int bi = 0; bi < DIGS; bi++) {  // Теперь пробегаем по цифрам (разрядам)
                                      // индикатора
    if(seg == 7 && bi == dpPos) // Подсветим нужную десятичную точку
      digitalWrite(digPins[bi], L_LOW);
    else {
      if(digit[d[bi]] & mask) // Если данный сегмент в bi-ом разряде должен
        digitalWrite(digPins[bi], L_LOW); // гореть, то уроним потенциал катода
										                      // соответствующей цифры (разряда)
      else // А если гореть не должен, то потенциал катода bi-й цифры
        digitalWrite(digPins[bi], L_HIGH); // должен быть высокий 
    }
  }
}
//-----------------------------------------------------------------------------
uint8_t GetKey(void) { // Возвращает состояние кнопки
  uint8_t ret, bcs = !(PINC & bpm); // Текущее состояние кнопки
  if(!bos && bcs) { // Произошло нажатие кнопки
    tpr = millis(); bos = bcs; // Фиксируем время первого контакта и состояние
    return NO_ACT;
  }
  if(bos && !bcs) { // Произошло либо отжатие или дребезг кнопки
    if((millis() - tpr) < dtBounce) return NO_ACT; // Дребезг, однако...
    if((millis() - tpr) < dtPress) { // Отжатие кнопки после короткого нажатия
      ret = bos; bos = bcs;
      return ret;
    }
    ret = bos << 4; bos = bcs; // Отжатие кнопки после ее длительного удержания
    return ret;
  }
  if(bos && bcs) { // Предыдущее и текущее состояние - нажатое. Удержание?
    if((millis() - tpr) < dtPress) return NO_ACT; // Но пока еще не ясно...
    ret = bos << 2; // И, все-таки, удержание - возвращаем соответствующий код
    return ret;
  }
  return NO_ACT;
}
//-----------------------------------------------------------------------------
#endif  // Конец блока констант, перемнных и функций расширенного варианта
//-----------------------------------------------------------------------------
int main(void) {
  // Инициализация таймера 2 для тактирования процессов с точностью 1 мс
  TCCR2A = (1<<WGM21);	// Режим CTC (сброс по совпадению)
  TCCR2B = (1<<CS22); 	// Предделитель 64
  OCR2A = 249;			// При этом значении прерывания происходят каждую мс
  TIMSK2 = (1<<OCIE2A);	// Разрешить прерывание по совпадению

  // Инициализация TWI
  TWAR = ADDR<<1; // Установка адреса (старшие 7 бит TWAR)
  TWCR = (1<<TWINT) | // Сбросим флаг TWINT
         (1<<TWEA)  | // Разрешаем подтверждение адреса
         (1<<TWEN)  | // Разрешаем работу TWI
         (1<<TWIE);

  sei(); // Разрешаем прерывания

  DDRD |= mask1; // Пины шин - на вывод
  PORTD |= mask1; // Отпускаем все шины

#ifndef LITE // Расширенный вариант (с индикатором и кнопкой)
  bpm = 0x01 << b_pin;	// Маска пина, к которому подключена кнопка
  DDRC &= ~bpm; 		// Пин порта C с кнопкой настроим на ввод
  PORTC |= bpm; 		// Активируем подтягивающий резистор на пине с кнопкой

  maxval = 1;
  for(int i = 0; i < DIGS; i++)
    maxval *= 10;

  for(int i = 0; i < 8; i++) { // Инициализация пинов сергемнтов
    pinMode(segPins[i], OUTPUT);
    digitalWrite(segPins[i], L_LOW);
  }
  for(int i = 0; i < DIGS; i++) { // Инициализация пинов цифр
    pinMode(digPins[i], OUTPUT);
    digitalWrite(digPins[i], L_HIGH);
  }
  // Проверим работу всех сегментов - покажем их
  val = 8888; UpdateDigits(val);
  uint32_t st = millis() + 1000;
  while(millis() < st) {}
#ifdef INDICATOR_TEST // Тест индикатора - показываем цифры в цикле...
  while(1) {
    int dt = (val > 1000)? 1 : (val > 100)? 10 : (val > 10)? 100 : 1000;
    dpPos = (val > 1000)? 0 : (val > 100)? 1 : (val > 10)? 2 : 3;
    UpdateDigits(val);
    st = millis() + dt;
    while(millis() < st) { }
    if(++val > 9999) val = 0;
  }
#endif // Конец теста индикатора
#ifdef BUTTON_TEST // Тест кнопки - давим по-разному, смотрим на индикатор
  int oval = 0;
  while(1) {
    switch(GetKey()) {
      case PLUS_CLICK: val = 1; break;
      case PLUS_PRESS: val = 2; break;
      case PLUS_PRESS_REL: val = 3; break;
    }
    if(oval != val) {
      UpdateDigits(val);
      oval = val;
    }
  }
#endif // Конец теста кнопки
#endif // Конец блока расширенного варианта

#ifdef DEBUG
  DDRC |= ledMask;
  PORTC &= ~ledMask;
#endif

  uint32_t stConv = millis(); // Для тактирования запуска датчиков
  uint32_t spConv;

  while(1) {
    if(millis() >= stConv) { // Раз в секунду запускаем датчики
//      PORTC |= ledMask; _delay_ms(100); PORTC &= ~ledMask;
      stConv = millis() + 1000;
      if(w1Status == W1_READY) { // Начинаем цикл преобразования
        devs = w1_reset() & mask1;
        if(devs) { // Датчики есть
          w1_write_byte(0xcc); // Skip ROM
          w1_write_byte(0x44); // Запускаем все датчики
          w1Status = W1_CONVERSION;
          spConv = millis() + 800;
        }
      }
    }
    if((w1Status == W1_CONVERSION) && (millis() >= spConv)) {
      TWCR &= ~(1<<TWEA); // Блокируем шину I2C на время обновления буфера
      w1_reset();
      w1_write_byte(0xcc); // Skip ROM
      w1_write_byte(0xbe); // Будем читать температуры
      uint8_t bytes[8]; // Вспомогательный буфер
      w1_read_bytes(bytes); // Старшие байты температурных кода
      for(uint8_t i = 0; i < 8; i++) // Занесем их в буфер
        buf[2*i + 1] = bytes[i];
      w1_read_bytes(bytes); // Младшие байты температурных кода
      for(uint8_t i = 0; i < 8; i++) // Занесем их в буфер
        buf[2*i] = bytes[i];
      w1_reset();
      w1Status = W1_READY;
      TWCR |= (1<<TWEA); // Разрешаем работу с шиной I2C
      gFlag = 1;
    }
#ifndef LITE // Расширенный вариант (с индикатором и кнопкой)
    switch(GetKey()) { // Анализируем состояние кнопки
      case PLUS_CLICK: // Короткое нажатие кнопки
        dMode = SHOW_TEMP; dpPos = 1;
        if(++curTS >= 4) curTS = 0;
        gFlag = 1;
        break;
      case PLUS_PRESS: // Удержание кнопки нажатой
        if(dMode != SHOW_TS_NUM) {
          dMode = SHOW_TS_NUM; dpPos = DIGS + 1;
          val = curTS;
          UpdateDigits(val);
        }
        break;
      case PLUS_PRESS_REL: // Кнопка отжата после удерживания
        dMode = SHOW_TEMP; dpPos = 1;
        gFlag = 1;
        break;
    }
    if(dMode == SHOW_TEMP && gFlag) {
      uint16_t code = (((uint16_t)buf[2*curTS]) << 8) | buf[2*curTS + 1];
      val = (uint16_t)(code*6.25 + 0.5);
      UpdateDigits(val); gFlag = 0;
    }
#endif // Конец расширенного варианта
  }
}
//-----------------------------------------------------------------------------

